@@ -15,17 +15,16 @@ MVMint64 MVM_6model_container_iscont_rw(MVMThreadContext *tc, MVMObject *cont) {
  * ***************************************************************************/
 
 typedef struct {
-    MVMObject *fetch_code;
-    MVMObject *store_code;
+    MVMCode *fetch_code;
+    MVMCode *store_code;
 } CodePairContData;
 
 static void code_pair_fetch_internal(MVMThreadContext *tc, MVMObject *cont, MVMRegister *res, MVMReturnType res_type) {
-    CodePairContData        *data = (CodePairContData *)STABLE(cont)->container_data;
-    MVMObject               *code = MVM_frame_find_invokee(tc, data->fetch_code, NULL);
-    MVMCallsite *inv_arg_callsite = MVM_callsite_get_common(tc, MVM_CALLSITE_ID_INV_ARG);
-    MVM_args_setup_thunk(tc, res, res_type, inv_arg_callsite);
-    tc->cur_frame->args[0].o      = cont;
-    STABLE(code)->invoke(tc, code, inv_arg_callsite, tc->cur_frame->args);
+    CodePairContData *data = (CodePairContData *)STABLE(cont)->container_data;
+    MVMCallStackArgsFromC *args_record = MVM_callstack_allocate_args_from_c(tc,
+            MVM_callsite_get_common(tc, MVM_CALLSITE_ID_OBJ));
+    args_record->args.source[0].o = cont;
+    MVM_frame_dispatch_from_c(tc, data->fetch_code, args_record, res, res_type);
 }
 
 static void code_pair_fetch(MVMThreadContext *tc, MVMObject *cont, MVMRegister *res) {
@@ -36,6 +35,10 @@ static void code_pair_fetch_i(MVMThreadContext *tc, MVMObject *cont, MVMRegister
     code_pair_fetch_internal(tc, cont, res, MVM_RETURN_INT);
 }
 
+static void code_pair_fetch_u(MVMThreadContext *tc, MVMObject *cont, MVMRegister *res) {
+    code_pair_fetch_internal(tc, cont, res, MVM_RETURN_UINT);
+}
+
 static void code_pair_fetch_n(MVMThreadContext *tc, MVMObject *cont, MVMRegister *res) {
     code_pair_fetch_internal(tc, cont, res, MVM_RETURN_NUM);
 }
@@ -45,18 +48,17 @@ static void code_pair_fetch_s(MVMThreadContext *tc, MVMObject *cont, MVMRegister
 }
 
 static void code_pair_store_internal(MVMThreadContext *tc, MVMObject *cont, MVMRegister value, MVMCallsite *cs) {
-    CodePairContData         *data = (CodePairContData *)STABLE(cont)->container_data;
-    MVMObject                *code = MVM_frame_find_invokee(tc, data->store_code, NULL);
-    MVM_args_setup_thunk(tc, NULL, MVM_RETURN_VOID, cs);
-    tc->cur_frame->args[0].o       = cont;
-    tc->cur_frame->args[1]         = value;
-    STABLE(code)->invoke(tc, code, cs, tc->cur_frame->args);
+    CodePairContData *data = (CodePairContData *)STABLE(cont)->container_data;
+    MVMCallStackArgsFromC *args_record = MVM_callstack_allocate_args_from_c(tc, cs);
+    args_record->args.source[0].o = cont;
+    args_record->args.source[1] = value;
+    MVM_frame_dispatch_from_c(tc, data->store_code, args_record, NULL, MVM_RETURN_VOID);
 }
 
 static void code_pair_store(MVMThreadContext *tc, MVMObject *cont, MVMObject *obj) {
     MVMRegister r;
     r.o = obj;
-    code_pair_store_internal(tc, cont, r, MVM_callsite_get_common(tc, MVM_CALLSITE_ID_TWO_OBJ));
+    code_pair_store_internal(tc, cont, r, MVM_callsite_get_common(tc, MVM_CALLSITE_ID_OBJ_OBJ));
 }
 
 static void code_pair_store_i(MVMThreadContext *tc, MVMObject *cont, MVMint64 value) {
@@ -79,7 +81,6 @@ static void code_pair_store_s(MVMThreadContext *tc, MVMObject *cont, MVMString *
 
 static void code_pair_gc_mark_data(MVMThreadContext *tc, MVMSTable *st, MVMGCWorklist *worklist) {
     CodePairContData *data = (CodePairContData *)st->container_data;
-
     MVM_gc_worklist_add(tc, worklist, &data->fetch_code);
     MVM_gc_worklist_add(tc, worklist, &data->store_code);
 }
@@ -90,14 +91,12 @@ static void code_pair_gc_free_data(MVMThreadContext *tc, MVMSTable *st) {
 
 static void code_pair_serialize(MVMThreadContext *tc, MVMSTable *st, MVMSerializationWriter *writer) {
     CodePairContData *data = (CodePairContData *)st->container_data;
-
-    MVM_serialization_write_ref(tc, writer, data->fetch_code);
-    MVM_serialization_write_ref(tc, writer, data->store_code);
+    MVM_serialization_write_ref(tc, writer, (MVMObject *)data->fetch_code);
+    MVM_serialization_write_ref(tc, writer, (MVMObject *)data->store_code);
 }
 
 static void code_pair_deserialize(MVMThreadContext *tc, MVMSTable *st, MVMSerializationReader *reader) {
     CodePairContData *data = (CodePairContData *)st->container_data;
-
     MVM_ASSIGN_REF(tc, &(st->header), data->fetch_code, MVM_serialization_read_ref(tc, reader));
     MVM_ASSIGN_REF(tc, &(st->header), data->store_code, MVM_serialization_read_ref(tc, reader));
 }
@@ -110,10 +109,12 @@ static const MVMContainerSpec code_pair_spec = {
     "code_pair",
     code_pair_fetch,
     code_pair_fetch_i,
+    code_pair_fetch_u,
     code_pair_fetch_n,
     code_pair_fetch_s,
     code_pair_store,
     code_pair_store_i,
+    (void *)code_pair_store_i, /* FIXME need a code_pair_store_u but lacking tests showing this need */
     code_pair_store_n,
     code_pair_store_s,
     code_pair_store,
@@ -125,14 +126,13 @@ static const MVMContainerSpec code_pair_spec = {
     NULL,
     code_pair_can_store,
     NULL, /* cas */
-    NULL, /* atomic_load */
-    NULL, /* atomic_store */
+    NULL, /* load_atomic */
+    NULL, /* store_atomic */
     0
 };
 
 static void code_pair_set_container_spec(MVMThreadContext *tc, MVMSTable *st) {
     CodePairContData *data = MVM_malloc(sizeof(CodePairContData));
-
     data->fetch_code   = NULL;
     data->store_code   = NULL;
     st->container_data = data;
@@ -141,26 +141,29 @@ static void code_pair_set_container_spec(MVMThreadContext *tc, MVMSTable *st) {
 
 static void code_pair_configure_container_spec(MVMThreadContext *tc, MVMSTable *st, MVMObject *config) {
     CodePairContData *data = (CodePairContData *)st->container_data;
-
     MVMROOT2(tc, config, st, {
         MVMString *fetch = MVM_string_ascii_decode_nt(tc, tc->instance->VMString, "fetch");
-        MVMString *store;
 
         if (!MVM_repr_exists_key(tc, config, fetch))
             MVM_exception_throw_adhoc(tc, "Container spec 'code_pair' must be configured with a fetch");
+        MVMObject *fetch_code = MVM_repr_at_key_o(tc, config, fetch);
+        if (!MVM_code_iscode(tc, fetch_code))
+            MVM_exception_throw_adhoc(tc, "Container spec 'code_pair' must be configured with a code handle");
+        MVM_ASSIGN_REF(tc, &(st->header), data->fetch_code, (MVMCode *)fetch_code);
 
-        MVM_ASSIGN_REF(tc, &(st->header), data->fetch_code, MVM_repr_at_key_o(tc, config, fetch));
-
-        store = MVM_string_ascii_decode_nt(tc, tc->instance->VMString, "store");
+        MVMString *store = MVM_string_ascii_decode_nt(tc, tc->instance->VMString, "store");
 
         if (!MVM_repr_exists_key(tc, config, store))
             MVM_exception_throw_adhoc(tc, "Container spec 'code_pair' must be configured with a store");
-
-        MVM_ASSIGN_REF(tc, &(st->header), data->store_code, MVM_repr_at_key_o(tc, config, store));
+        MVMObject *store_code = MVM_repr_at_key_o(tc, config, store);
+        if (!MVM_code_iscode(tc, store_code))
+            MVM_exception_throw_adhoc(tc, "Container spec 'code_pair' must be configured with a code handle");
+        MVM_ASSIGN_REF(tc, &(st->header), data->store_code, store_code);
     });
 }
 
-static const MVMContainerConfigurer CodePairContainerConfigurer = {
+static MVMContainerConfigurer CodePairContainerConfigurer = {
+    NULL,
     code_pair_set_container_spec,
     code_pair_configure_container_spec
 };
@@ -178,12 +181,12 @@ typedef struct {
     size_t descriptor_offset;
 
     /* Callbacks. */
-    MVMObject *store;
-    MVMObject *store_unchecked;
-    MVMObject *cas;
-    MVMObject *atomic_store;
+    MVMCode *store;
+    MVMCode *store_unchecked;
+    MVMCode *cas;
+    MVMCode *store_atomic;
 
-    /* Retained for serializatin purposes only. */
+    /* Retained for serialization purposes only. */
     MVMObject *attrs_class;
     MVMString *value_attr;
     MVMString *descriptor_attr;
@@ -215,6 +218,10 @@ static void value_desc_cont_fetch_i(MVMThreadContext *tc, MVMObject *cont, MVMRe
     res->i64 = MVM_repr_get_int(tc, read_container_value(cont));
 }
 
+static void value_desc_cont_fetch_u(MVMThreadContext *tc, MVMObject *cont, MVMRegister *res) {
+    res->u64 = MVM_repr_get_uint(tc, read_container_value(cont));
+}
+
 static void value_desc_cont_fetch_n(MVMThreadContext *tc, MVMObject *cont, MVMRegister *res) {
     res->n64 = MVM_repr_get_num(tc, read_container_value(cont));
 }
@@ -225,12 +232,11 @@ static void value_desc_cont_fetch_s(MVMThreadContext *tc, MVMObject *cont, MVMRe
 
 static void value_desc_cont_store(MVMThreadContext *tc, MVMObject *cont, MVMObject *value) {
     MVMValueDescContainer *data = (MVMValueDescContainer *)STABLE(cont)->container_data;
-    MVMObject *code = MVM_frame_find_invokee(tc, data->store, NULL);
-    MVMCallsite *cs = MVM_callsite_get_common(tc, MVM_CALLSITE_ID_TWO_OBJ);
-    MVM_args_setup_thunk(tc, NULL, MVM_RETURN_VOID, cs);
-    tc->cur_frame->args[0].o = cont;
-    tc->cur_frame->args[1].o = value;
-    STABLE(code)->invoke(tc, code, cs, tc->cur_frame->args);
+    MVMCallStackArgsFromC *args_record = MVM_callstack_allocate_args_from_c(tc,
+            MVM_callsite_get_common(tc, MVM_CALLSITE_ID_OBJ_OBJ));
+    args_record->args.source[0].o = cont;
+    args_record->args.source[1].o = value;
+    MVM_frame_dispatch_from_c(tc, data->store, args_record, NULL, MVM_RETURN_VOID);
 }
 
 static void value_desc_cont_store_i(MVMThreadContext *tc, MVMObject *cont, MVMint64 value) {
@@ -259,12 +265,11 @@ static void value_desc_cont_store_s(MVMThreadContext *tc, MVMObject *cont, MVMSt
 
 static void value_desc_cont_store_unchecked(MVMThreadContext *tc, MVMObject *cont, MVMObject *value) {
     MVMValueDescContainer *data = (MVMValueDescContainer *)STABLE(cont)->container_data;
-    MVMObject *code = MVM_frame_find_invokee(tc, data->store_unchecked, NULL);
-    MVMCallsite *cs = MVM_callsite_get_common(tc, MVM_CALLSITE_ID_TWO_OBJ);
-    MVM_args_setup_thunk(tc, NULL, MVM_RETURN_VOID, cs);
-    tc->cur_frame->args[0].o = cont;
-    tc->cur_frame->args[1].o = value;
-    STABLE(code)->invoke(tc, code, cs, tc->cur_frame->args);
+    MVMCallStackArgsFromC *args_record = MVM_callstack_allocate_args_from_c(tc,
+            MVM_callsite_get_common(tc, MVM_CALLSITE_ID_OBJ_OBJ));
+    args_record->args.source[0].o = cont;
+    args_record->args.source[1].o = value;
+    MVM_frame_dispatch_from_c(tc, data->store_unchecked, args_record, NULL, MVM_RETURN_VOID);
 }
 
 static void value_desc_cont_gc_mark_data(MVMThreadContext *tc, MVMSTable *st, MVMGCWorklist *worklist) {
@@ -272,7 +277,7 @@ static void value_desc_cont_gc_mark_data(MVMThreadContext *tc, MVMSTable *st, MV
     MVM_gc_worklist_add(tc, worklist, &data->store);
     MVM_gc_worklist_add(tc, worklist, &data->store_unchecked);
     MVM_gc_worklist_add(tc, worklist, &data->cas);
-    MVM_gc_worklist_add(tc, worklist, &data->atomic_store);
+    MVM_gc_worklist_add(tc, worklist, &data->store_atomic);
     MVM_gc_worklist_add(tc, worklist, &data->attrs_class);
     MVM_gc_worklist_add(tc, worklist, &data->value_attr);
     MVM_gc_worklist_add(tc, worklist, &data->descriptor_attr);
@@ -284,10 +289,10 @@ static void value_desc_cont_gc_free_data(MVMThreadContext *tc, MVMSTable *st) {
 
 static void value_desc_cont_serialize(MVMThreadContext *tc, MVMSTable *st, MVMSerializationWriter *writer) {
     MVMValueDescContainer *data = (MVMValueDescContainer *)st->container_data;
-    MVM_serialization_write_ref(tc, writer, data->store);
-    MVM_serialization_write_ref(tc, writer, data->store_unchecked);
-    MVM_serialization_write_ref(tc, writer, data->cas);
-    MVM_serialization_write_ref(tc, writer, data->atomic_store);
+    MVM_serialization_write_ref(tc, writer, (MVMObject *)data->store);
+    MVM_serialization_write_ref(tc, writer, (MVMObject *)data->store_unchecked);
+    MVM_serialization_write_ref(tc, writer, (MVMObject *)data->cas);
+    MVM_serialization_write_ref(tc, writer, (MVMObject *)data->store_atomic);
     MVM_serialization_write_ref(tc, writer, data->attrs_class);
     MVM_serialization_write_str(tc, writer, data->value_attr);
     MVM_serialization_write_str(tc, writer, data->descriptor_attr);
@@ -298,7 +303,7 @@ static void value_desc_cont_deserialize(MVMThreadContext *tc, MVMSTable *st, MVM
     MVM_ASSIGN_REF(tc, &(st->header), data->store, MVM_serialization_read_ref(tc, reader));
     MVM_ASSIGN_REF(tc, &(st->header), data->store_unchecked, MVM_serialization_read_ref(tc, reader));
     MVM_ASSIGN_REF(tc, &(st->header), data->cas, MVM_serialization_read_ref(tc, reader));
-    MVM_ASSIGN_REF(tc, &(st->header), data->atomic_store, MVM_serialization_read_ref(tc, reader));
+    MVM_ASSIGN_REF(tc, &(st->header), data->store_atomic, MVM_serialization_read_ref(tc, reader));
     MVM_ASSIGN_REF(tc, &(st->header), data->attrs_class, MVM_serialization_read_ref(tc, reader));
     MVM_ASSIGN_REF(tc, &(st->header), data->value_attr, MVM_serialization_read_str(tc, reader));
     MVM_ASSIGN_REF(tc, &(st->header), data->descriptor_attr, MVM_serialization_read_str(tc, reader));
@@ -334,13 +339,12 @@ static void value_desc_cont_cas(MVMThreadContext *tc, MVMObject *cont,
                               MVMObject *expected, MVMObject *value,
                               MVMRegister *result) {
     MVMValueDescContainer *data = (MVMValueDescContainer *)STABLE(cont)->container_data;
-    MVMObject *code = MVM_frame_find_invokee(tc, data->cas, NULL);
-    MVMCallsite *cs = MVM_callsite_get_common(tc, MVM_CALLSITE_ID_TYPECHECK);
-    MVM_args_setup_thunk(tc, result, MVM_RETURN_OBJ, cs);
-    tc->cur_frame->args[0].o = cont;
-    tc->cur_frame->args[1].o = expected;
-    tc->cur_frame->args[2].o = value;
-    STABLE(code)->invoke(tc, code, cs, tc->cur_frame->args);
+    MVMCallStackArgsFromC *args_record = MVM_callstack_allocate_args_from_c(tc,
+            MVM_callsite_get_common(tc, MVM_CALLSITE_ID_OBJ_OBJ_OBJ));
+    args_record->args.source[0].o = cont;
+    args_record->args.source[1].o = expected;
+    args_record->args.source[2].o = value;
+    MVM_frame_dispatch_from_c(tc, data->cas, args_record, result, MVM_RETURN_OBJ);
 }
 
 static MVMObject * value_desc_cont_atomic_load(MVMThreadContext *tc, MVMObject *cont) {
@@ -349,24 +353,25 @@ static MVMObject * value_desc_cont_atomic_load(MVMThreadContext *tc, MVMObject *
     return value ? value : tc->instance->VMNull;
 }
 
-void value_desc_cont_atomic_store(MVMThreadContext *tc, MVMObject *cont, MVMObject *value) {
+static void value_desc_cont_atomic_store(MVMThreadContext *tc, MVMObject *cont, MVMObject *value) {
     MVMValueDescContainer *data = (MVMValueDescContainer *)STABLE(cont)->container_data;
-    MVMObject *code = MVM_frame_find_invokee(tc, data->atomic_store, NULL);
-    MVMCallsite *cs = MVM_callsite_get_common(tc, MVM_CALLSITE_ID_TWO_OBJ);
-    MVM_args_setup_thunk(tc, NULL, MVM_RETURN_VOID, cs);
-    tc->cur_frame->args[0].o = cont;
-    tc->cur_frame->args[1].o = value;
-    STABLE(code)->invoke(tc, code, cs, tc->cur_frame->args);
+    MVMCallStackArgsFromC *args_record = MVM_callstack_allocate_args_from_c(tc,
+            MVM_callsite_get_common(tc, MVM_CALLSITE_ID_OBJ_OBJ));
+    args_record->args.source[0].o = cont;
+    args_record->args.source[1].o = value;
+    MVM_frame_dispatch_from_c(tc, data->store_atomic, args_record, NULL, MVM_RETURN_VOID);
 }
 
 static const MVMContainerSpec value_desc_cont_spec = {
     "value_desc_cont",
     value_desc_cont_fetch,
     value_desc_cont_fetch_i,
+    value_desc_cont_fetch_u,
     value_desc_cont_fetch_n,
     value_desc_cont_fetch_s,
     value_desc_cont_store,
     value_desc_cont_store_i,
+    (void *)value_desc_cont_store_i, /* FIXME need a value_desc_cont_store_u but lacking tests showing this need */
     value_desc_cont_store_n,
     value_desc_cont_store_s,
     value_desc_cont_store_unchecked,
@@ -405,13 +410,21 @@ static void value_desc_cont_configure_container_spec(MVMThreadContext *tc, MVMST
     MVMROOT2(tc, st, config, {
         MVMObject *value;
         value = grab_one_value(tc, config, "store");
+        if (!MVM_code_iscode(tc, value))
+            MVM_exception_throw_adhoc(tc, "Container spec must be configured with a code handle");
         MVM_ASSIGN_REF(tc, &(st->header), data->store, value);
         value = grab_one_value(tc, config, "store_unchecked");
+        if (!MVM_code_iscode(tc, value))
+            MVM_exception_throw_adhoc(tc, "Container spec must be configured with a code handle");
         MVM_ASSIGN_REF(tc, &(st->header), data->store_unchecked, value);
         value = grab_one_value(tc, config, "cas");
+        if (!MVM_code_iscode(tc, value))
+            MVM_exception_throw_adhoc(tc, "Container spec must be configured with a code handle");
         MVM_ASSIGN_REF(tc, &(st->header), data->cas, value);
         value = grab_one_value(tc, config, "atomic_store");
-        MVM_ASSIGN_REF(tc, &(st->header), data->atomic_store, value);
+        if (!MVM_code_iscode(tc, value))
+            MVM_exception_throw_adhoc(tc, "Container spec must be configured with a code handle");
+        MVM_ASSIGN_REF(tc, &(st->header), data->store_atomic, value);
         value = grab_one_value(tc, config, "attrs_class");
         MVM_ASSIGN_REF(tc, &(st->header), data->attrs_class, value);
         value = grab_one_value(tc, config, "value_attr");
@@ -422,7 +435,8 @@ static void value_desc_cont_configure_container_spec(MVMThreadContext *tc, MVMST
     calculate_attr_offsets(tc, st, data);
 }
 
-static const MVMContainerConfigurer ValueDescContainerConfigurer = {
+static MVMContainerConfigurer ValueDescContainerConfigurer = {
+    NULL,
     value_desc_cont_set_container_spec,
     value_desc_cont_configure_container_spec
 };
@@ -453,6 +467,28 @@ static void native_ref_fetch_i(MVMThreadContext *tc, MVMObject *cont, MVMRegiste
     }
 }
 
+static void native_ref_fetch_u(MVMThreadContext *tc, MVMObject *cont, MVMRegister *res) {
+    MVMNativeRefREPRData *repr_data = (MVMNativeRefREPRData *)STABLE(cont)->REPR_data;
+    if (repr_data->primitive_type != MVM_STORAGE_SPEC_BP_UINT64)
+        MVM_exception_throw_adhoc(tc, "This container does not reference a native unsigned integer");
+    switch (repr_data->ref_kind) {
+        case MVM_NATIVEREF_LEX:
+            res->u64 = MVM_nativeref_read_lex_i(tc, cont); /* covers unsigned as well */
+            break;
+        case MVM_NATIVEREF_ATTRIBUTE:
+            res->u64 = MVM_nativeref_read_attribute_u(tc, cont);
+            break;
+        case MVM_NATIVEREF_POSITIONAL:
+            res->u64 = MVM_nativeref_read_positional_u(tc, cont);
+            break;
+        case MVM_NATIVEREF_MULTIDIM:
+            res->u64 = MVM_nativeref_read_multidim_u(tc, cont);
+            break;
+        default:
+            MVM_exception_throw_adhoc(tc, "Unknown native int reference kind");
+    }
+}
+
 static void native_ref_fetch_n(MVMThreadContext *tc, MVMObject *cont, MVMRegister *res) {
     MVMNativeRefREPRData *repr_data = (MVMNativeRefREPRData *)STABLE(cont)->REPR_data;
     if (repr_data->primitive_type != MVM_STORAGE_SPEC_BP_NUM)
@@ -505,8 +541,15 @@ static void native_ref_fetch(MVMThreadContext *tc, MVMObject *cont, MVMRegister
         hll = MVM_hll_current(tc);
     switch (repr_data->primitive_type) {
         case MVM_STORAGE_SPEC_BP_INT:
-            native_ref_fetch_i(tc, cont, &tmp);
-            res->o = MVM_repr_box_int(tc, hll->int_box_type, tmp.i64);
+        case MVM_STORAGE_SPEC_BP_UINT64:
+            if (repr_data->is_unsigned) {
+                native_ref_fetch_u(tc, cont, &tmp);
+                res->o = MVM_repr_box_uint(tc, hll->int_box_type, tmp.u64);
+            }
+            else {
+                native_ref_fetch_i(tc, cont, &tmp);
+                res->o = MVM_repr_box_int(tc, hll->int_box_type, tmp.i64);
+            }
             break;
         case MVM_STORAGE_SPEC_BP_NUM:
             native_ref_fetch_n(tc, cont, &tmp);
@@ -543,6 +586,28 @@ static void native_ref_store_i(MVMThreadContext *tc, MVMObject *cont, MVMint64 v
     }
 }
 
+static void native_ref_store_u(MVMThreadContext *tc, MVMObject *cont, MVMuint64 value) {
+    MVMNativeRefREPRData *repr_data = (MVMNativeRefREPRData *)STABLE(cont)->REPR_data;
+    if (repr_data->primitive_type != MVM_STORAGE_SPEC_BP_UINT64)
+        MVM_exception_throw_adhoc(tc, "This container does not reference a native integer");
+    switch (repr_data->ref_kind) {
+        case MVM_NATIVEREF_LEX:
+            MVM_nativeref_write_lex_u(tc, cont, value);
+            break;
+        case MVM_NATIVEREF_ATTRIBUTE:
+            MVM_nativeref_write_attribute_u(tc, cont, value);
+            break;
+        case MVM_NATIVEREF_POSITIONAL:
+            MVM_nativeref_write_positional_u(tc, cont, value);
+            break;
+        case MVM_NATIVEREF_MULTIDIM:
+            MVM_nativeref_write_multidim_u(tc, cont, value);
+            break;
+        default:
+            MVM_exception_throw_adhoc(tc, "Unknown native int reference kind");
+    }
+}
+
 static void native_ref_store_n(MVMThreadContext *tc, MVMObject *cont, MVMnum64 value) {
     MVMNativeRefREPRData *repr_data = (MVMNativeRefREPRData *)STABLE(cont)->REPR_data;
     if (repr_data->primitive_type != MVM_STORAGE_SPEC_BP_NUM)
@@ -590,8 +655,12 @@ static void native_ref_store_s(MVMThreadContext *tc, MVMObject *cont, MVMString
 static void native_ref_store(MVMThreadContext *tc, MVMObject *cont, MVMObject *obj) {
     MVMNativeRefREPRData *repr_data = (MVMNativeRefREPRData *)STABLE(cont)->REPR_data;
     switch (repr_data->primitive_type) {
+        case MVM_STORAGE_SPEC_BP_UINT64:
         case MVM_STORAGE_SPEC_BP_INT:
-            native_ref_store_i(tc, cont, MVM_repr_get_int(tc, obj));
+            if (repr_data->is_unsigned)
+                native_ref_store_u(tc, cont, MVM_repr_get_uint(tc, obj));
+            else
+                native_ref_store_i(tc, cont, MVM_repr_get_int(tc, obj));
             break;
         case MVM_STORAGE_SPEC_BP_NUM:
             native_ref_store_n(tc, cont, MVM_repr_get_num(tc, obj));
@@ -620,10 +689,12 @@ static const MVMContainerSpec native_ref_spec = {
     "native_ref",
     native_ref_fetch,
     native_ref_fetch_i,
+    native_ref_fetch_u,
     native_ref_fetch_n,
     native_ref_fetch_s,
     native_ref_store,
     native_ref_store_i,
+    native_ref_store_u,
     native_ref_store_n,
     native_ref_store_s,
     native_ref_store,
@@ -635,8 +706,8 @@ static const MVMContainerSpec native_ref_spec = {
     NULL,
     native_ref_can_store,
     NULL, /* cas */
-    NULL, /* atomic_load */
-    NULL, /* atomic_store */
+    NULL, /* load_atomic */
+    NULL, /* store_atomic */
     1
 };
 
@@ -649,14 +720,13 @@ static void native_ref_configure_container_spec(MVMThreadContext *tc, MVMSTable
 }
 
 void *MVM_container_devirtualize_fetch_for_jit(MVMThreadContext *tc, MVMSTable *st, MVMuint16 type) {
-    if (type != MVM_reg_int64)
-        return NULL;
     if (st->container_spec == &native_ref_spec) {
         MVMNativeRefREPRData *repr_data = (MVMNativeRefREPRData *)st->REPR_data;
         switch (repr_data->ref_kind) {
             case MVM_NATIVEREF_LEX:
                 switch (type) {
                     case MVM_reg_int64: return MVM_nativeref_read_lex_i;
+                    case MVM_reg_uint64: return MVM_nativeref_read_lex_i;
                     case MVM_reg_num64: return MVM_nativeref_read_lex_n;
                     case MVM_reg_str:   return MVM_nativeref_read_lex_s;
                 }
@@ -664,6 +734,7 @@ void *MVM_container_devirtualize_fetch_for_jit(MVMThreadContext *tc, MVMSTable *
             case MVM_NATIVEREF_ATTRIBUTE:
                 switch (type) {
                     case MVM_reg_int64: return MVM_nativeref_read_attribute_i;
+                    case MVM_reg_uint64: return MVM_nativeref_read_attribute_u;
                     case MVM_reg_num64: return MVM_nativeref_read_attribute_n;
                     case MVM_reg_str:   return MVM_nativeref_read_attribute_s;
                 }
@@ -671,6 +742,7 @@ void *MVM_container_devirtualize_fetch_for_jit(MVMThreadContext *tc, MVMSTable *
             case MVM_NATIVEREF_POSITIONAL:
                 switch (type) {
                     case MVM_reg_int64: return MVM_nativeref_read_positional_i;
+                    case MVM_reg_uint64: return MVM_nativeref_read_positional_u;
                     case MVM_reg_num64: return MVM_nativeref_read_positional_n;
                     case MVM_reg_str:   return MVM_nativeref_read_positional_s;
                 }
@@ -678,6 +750,7 @@ void *MVM_container_devirtualize_fetch_for_jit(MVMThreadContext *tc, MVMSTable *
             case MVM_NATIVEREF_MULTIDIM:
                 switch (type) {
                     case MVM_reg_int64: return MVM_nativeref_read_multidim_i;
+                    case MVM_reg_uint64: return MVM_nativeref_read_multidim_u;
                     case MVM_reg_num64: return MVM_nativeref_read_multidim_n;
                     case MVM_reg_str:   return MVM_nativeref_read_multidim_s;
                 }
@@ -690,14 +763,13 @@ void *MVM_container_devirtualize_fetch_for_jit(MVMThreadContext *tc, MVMSTable *
 }
 
 void *MVM_container_devirtualize_store_for_jit(MVMThreadContext *tc, MVMSTable *st, MVMuint16 type) {
-    if (type != MVM_reg_int64)
-        return NULL;
     if (st->container_spec == &native_ref_spec) {
         MVMNativeRefREPRData *repr_data = (MVMNativeRefREPRData *)st->REPR_data;
         switch (repr_data->ref_kind) {
             case MVM_NATIVEREF_LEX:
                 switch (type) {
                     case MVM_reg_int64: return MVM_nativeref_write_lex_i;
+                    case MVM_reg_uint64: return MVM_nativeref_write_lex_u;
                     case MVM_reg_num64: return MVM_nativeref_write_lex_n;
                     case MVM_reg_str:   return MVM_nativeref_write_lex_s;
                 }
@@ -705,6 +777,7 @@ void *MVM_container_devirtualize_store_for_jit(MVMThreadContext *tc, MVMSTable *
             case MVM_NATIVEREF_ATTRIBUTE:
                 switch (type) {
                     case MVM_reg_int64: return MVM_nativeref_write_attribute_i;
+                    case MVM_reg_uint64: return MVM_nativeref_write_attribute_u;
                     case MVM_reg_num64: return MVM_nativeref_write_attribute_n;
                     case MVM_reg_str:   return MVM_nativeref_write_attribute_s;
                 }
@@ -712,6 +785,7 @@ void *MVM_container_devirtualize_store_for_jit(MVMThreadContext *tc, MVMSTable *
             case MVM_NATIVEREF_POSITIONAL:
                 switch (type) {
                     case MVM_reg_int64: return MVM_nativeref_write_positional_i;
+                    case MVM_reg_uint64: return MVM_nativeref_write_positional_u;
                     case MVM_reg_num64: return MVM_nativeref_write_positional_n;
                     case MVM_reg_str:   return MVM_nativeref_write_positional_s;
                 }
@@ -719,6 +793,7 @@ void *MVM_container_devirtualize_store_for_jit(MVMThreadContext *tc, MVMSTable *
             case MVM_NATIVEREF_MULTIDIM:
                 switch (type) {
                     case MVM_reg_int64: return MVM_nativeref_write_multidim_i;
+                    case MVM_reg_uint64: return MVM_nativeref_write_multidim_u;
                     case MVM_reg_num64: return MVM_nativeref_write_multidim_n;
                     case MVM_reg_str:   return MVM_nativeref_write_multidim_s;
                 }
@@ -730,7 +805,8 @@ void *MVM_container_devirtualize_store_for_jit(MVMThreadContext *tc, MVMSTable *
     return NULL;
 }
 
-static const MVMContainerConfigurer NativeRefContainerConfigurer = {
+static MVMContainerConfigurer NativeRefContainerConfigurer = {
+    NULL,
     native_ref_set_container_spec,
     native_ref_configure_container_spec
 };
@@ -740,23 +816,14 @@ static const MVMContainerConfigurer NativeRefContainerConfigurer = {
  * ***************************************************************************/
 
 /* Adds a container configurer to the registry. */
-void MVM_6model_add_container_config(MVMThreadContext *tc, MVMString *name,
-        const MVMContainerConfigurer *configurer) {
+static void add_container_config(MVMThreadContext *tc, const char *c_name,
+        MVMContainerConfigurer *configurer) {
+    MVMString *name = MVM_string_ascii_decode_nt(tc, tc->instance->VMString, c_name);
+    configurer->name = name;
 
-    if (!MVM_str_hash_key_is_valid(tc, name)) {
-        MVM_str_hash_key_throw_invalid(tc, name);
-    }
-
-    uv_mutex_lock(&tc->instance->mutex_container_registry);
-
-    MVMContainerRegistry *entry = MVM_str_hash_lvalue_fetch_nocheck(tc, &tc->instance->container_registry, name);
-
-    if (!entry->hash_handle.key) {
-        entry->configurer      = configurer;
-        entry->hash_handle.key = name;
-    }
-
-    uv_mutex_unlock(&tc->instance->mutex_container_registry);
+    MVMContainerConfigurer **target = MVM_fixkey_hash_insert_nocheck(tc, &tc->instance->container_registry, name);
+    *target = configurer;
+    MVM_gc_root_add_permanent_desc(tc, (MVMCollectable **)&configurer->name, c_name);
 }
 
 /* Gets a container configurer from the registry. */
@@ -766,22 +833,16 @@ const MVMContainerConfigurer * MVM_6model_get_container_config(MVMThreadContext
         MVM_str_hash_key_throw_invalid(tc, name);
     }
 
-    uv_mutex_lock(&tc->instance->mutex_container_registry);
-    MVMContainerRegistry *entry = MVM_str_hash_fetch_nocheck(tc, &tc->instance->container_registry, name);
-    uv_mutex_unlock(&tc->instance->mutex_container_registry);
-    return entry ? entry->configurer : NULL;
+    return MVM_fixkey_hash_fetch_nocheck(tc, &tc->instance->container_registry, name);
 }
 
 /* Does initial setup work of the container registry, including registering
  * the various built-in container types. */
 void MVM_6model_containers_setup(MVMThreadContext *tc) {
     /* Add built-in configurations. */
-    MVM_6model_add_container_config(tc,
-        MVM_string_ascii_decode_nt(tc, tc->instance->VMString, "code_pair"), &CodePairContainerConfigurer);
-    MVM_6model_add_container_config(tc,
-        MVM_string_ascii_decode_nt(tc, tc->instance->VMString, "native_ref"), &NativeRefContainerConfigurer);
-    MVM_6model_add_container_config(tc,
-        MVM_string_ascii_decode_nt(tc, tc->instance->VMString, "value_desc_cont"), &ValueDescContainerConfigurer);
+    add_container_config(tc, "code_pair", &CodePairContainerConfigurer);
+    add_container_config(tc, "native_ref", &NativeRefContainerConfigurer);
+    add_container_config(tc, "value_desc_cont", &ValueDescContainerConfigurer);
 }
 
 /* ***************************************************************************
@@ -798,7 +859,24 @@ static MVMint64 get_container_primitive(MVMThreadContext *tc, MVMObject *cont) {
     return MVM_STORAGE_SPEC_BP_NONE;
 }
 MVMint64 MVM_6model_container_iscont_i(MVMThreadContext *tc, MVMObject *cont) {
-    return get_container_primitive(tc, cont) == MVM_STORAGE_SPEC_BP_INT;
+    if (cont && IS_CONCRETE(cont)) {
+        const MVMContainerSpec *cs = STABLE(cont)->container_spec;
+        if (cs == &native_ref_spec && REPR(cont)->ID == MVM_REPR_ID_NativeRef) {
+            MVMNativeRefREPRData *repr_data = (MVMNativeRefREPRData *)STABLE(cont)->REPR_data;
+            return repr_data->primitive_type == MVM_STORAGE_SPEC_BP_INT && !repr_data->is_unsigned;
+        }
+    }
+    return 0;
+}
+MVMint64 MVM_6model_container_iscont_u(MVMThreadContext *tc, MVMObject *cont) {
+    if (cont && IS_CONCRETE(cont)) {
+        const MVMContainerSpec *cs = STABLE(cont)->container_spec;
+        if (cs == &native_ref_spec && REPR(cont)->ID == MVM_REPR_ID_NativeRef) {
+            MVMNativeRefREPRData *repr_data = (MVMNativeRefREPRData *)STABLE(cont)->REPR_data;
+            return repr_data->primitive_type == MVM_STORAGE_SPEC_BP_UINT64 && repr_data->is_unsigned;
+        }
+    }
+    return 0;
 }
 MVMint64 MVM_6model_container_iscont_n(MVMThreadContext *tc, MVMObject *cont) {
     return get_container_primitive(tc, cont) == MVM_STORAGE_SPEC_BP_NUM;
@@ -837,13 +915,12 @@ void MVM_6model_container_decont_s(MVMThreadContext *tc, MVMObject *cont, MVMReg
         res->s = MVM_repr_get_str(tc, cont);
 }
 
-/* If it's a container, do a fetch_i. Otherwise, try to unbox the received
+/* If it's a container, do a fetch_u. Otherwise, try to unbox the received
  * value as a native unsigned integer. */
 void MVM_6model_container_decont_u(MVMThreadContext *tc, MVMObject *cont, MVMRegister *res) {
     const MVMContainerSpec *cs = STABLE(cont)->container_spec;
     if (cs && IS_CONCRETE(cont))
-        /* XXX We need a fetch_u at some point. */
-        cs->fetch_i(tc, cont, res);
+        cs->fetch_u(tc, cont, res);
     else
         res->u64 = MVM_repr_get_uint(tc, cont);
 }
@@ -857,6 +934,15 @@ void MVM_6model_container_assign_i(MVMThreadContext *tc, MVMObject *cont, MVMint
         MVM_exception_throw_adhoc(tc, "Cannot assign to an immutable value");
 }
 
+/* Checks we have a container, and provided we do, assigns an uint into it. */
+void MVM_6model_container_assign_u(MVMThreadContext *tc, MVMObject *cont, MVMuint64 value) {
+    const MVMContainerSpec *cs = STABLE(cont)->container_spec;
+    if (cs && IS_CONCRETE(cont))
+        cs->store_u(tc, cont, value);
+    else
+        MVM_exception_throw_adhoc(tc, "Cannot assign to an immutable value");
+}
+
 /* Checks we have a container, and provided we do, assigns a num into it. */
 void MVM_6model_container_assign_n(MVMThreadContext *tc, MVMObject *cont, MVMnum64 value) {
     const MVMContainerSpec *cs = STABLE(cont)->container_spec;
@@ -909,8 +995,8 @@ MVMObject * MVM_6model_container_atomic_load(MVMThreadContext *tc, MVMObject *co
     if (IS_CONCRETE(cont)) {
         MVMContainerSpec const *cs = cont->st->container_spec;
         if (cs) {
-            if (cs->atomic_load)
-                return cs->atomic_load(tc, cont);
+            if (cs->load_atomic)
+                return cs->load_atomic(tc, cont);
             else
                 MVM_exception_throw_adhoc(tc,
                     "A %s container does not know how to do an atomic load",
@@ -933,8 +1019,8 @@ void MVM_6model_container_atomic_store(MVMThreadContext *tc, MVMObject *cont, MV
     if (IS_CONCRETE(cont)) {
         MVMContainerSpec const *cs = cont->st->container_spec;
         if (cs) {
-            if (cs->atomic_store)
-                cs->atomic_store(tc, cont, value);
+            if (cs->store_atomic)
+                cs->store_atomic(tc, cont, value);
             else
                 MVM_exception_throw_adhoc(tc,
                     "A %s container does not know how to do an atomic store",
@@ -977,7 +1063,7 @@ static AO_t * native_ref_as_atomic_i(MVMThreadContext *tc, MVMObject *cont) {
 
 MVMint64 MVM_6model_container_cas_i(MVMThreadContext *tc, MVMObject *cont,
                                     MVMint64 expected, MVMint64 value) {
-    return (MVMint64)MVM_cas(native_ref_as_atomic_i(tc, cont), (AO_t)expected, (AO_t)value);
+    return (MVMint64)MVM_cas(native_ref_as_atomic_i(tc, cont), AO_CAST(expected), AO_CAST(value));
 }
 
 MVMint64 MVM_6model_container_atomic_load_i(MVMThreadContext *tc, MVMObject *cont) {
@@ -997,5 +1083,5 @@ MVMint64 MVM_6model_container_atomic_dec(MVMThreadContext *tc, MVMObject *cont)
 }
 
 MVMint64 MVM_6model_container_atomic_add(MVMThreadContext *tc, MVMObject *cont, MVMint64 value) {
-    return (MVMint64)MVM_add(native_ref_as_atomic_i(tc, cont), (AO_t)value);
+    return (MVMint64)MVM_add(native_ref_as_atomic_i(tc, cont), AO_CAST(value));
 }
