@@ -14,13 +14,13 @@ static MVMint64 mp_get_int64(MVMThreadContext *tc, mp_int * a) {
      * but the negative max is -(2**63), which is 64 bits. */
     if (MP_NEG == a->sign) {
         if (bits > 64) {
-            MVM_exception_throw_adhoc(tc, "Cannot unbox %d bit wide bigint into native integer", bits);
+            MVM_exception_throw_adhoc(tc, "Cannot unbox %d bit wide bigint into native integer. Did you mix int and Int or literals?", bits);
         }
         ++signed_max;
     }
     else {
         if (bits > 63) {
-            MVM_exception_throw_adhoc(tc, "Cannot unbox %d bit wide bigint into native integer", bits);
+            MVM_exception_throw_adhoc(tc, "Cannot unbox %d bit wide bigint into native integer. Did you mix int and Int or literals?", bits);
         }
     }
 
@@ -28,7 +28,7 @@ static MVMint64 mp_get_int64(MVMThreadContext *tc, mp_int * a) {
 
     if (res > signed_max) {
         /* The mp_int was bigger than a signed result could be. */
-        MVM_exception_throw_adhoc(tc, "Cannot unbox %d bit wide bigint into native integer", bits);
+        MVM_exception_throw_adhoc(tc, "Cannot unbox %d bit wide bigint into native integer. Did you mix int and Int or literals?", bits);
     }
 
     return MP_NEG == a->sign ? -res : res;
@@ -43,7 +43,7 @@ static MVMuint64 mp_get_uint64(MVMThreadContext *tc, mp_int * a) {
     const int bits = mp_count_bits(a);
 
     if (bits > 64) {
-        MVM_exception_throw_adhoc(tc, "Cannot unbox %d bit wide bigint into native integer", bits);
+        MVM_exception_throw_adhoc(tc, "Cannot unbox %d bit wide bigint into native integer. Did you mix int and Int or literals?", bits);
     }
 
     return mp_get_mag_ull(a);
@@ -80,6 +80,7 @@ static void copy_to(MVMThreadContext *tc, MVMSTable *st, void *src, MVMObject *d
         mp_err err;
         dest_body->u.bigint = MVM_malloc(sizeof(mp_int));
         if ((err = mp_init_copy(dest_body->u.bigint, src_body->u.bigint)) != MP_OKAY) {
+            MVM_free(dest_body->u.bigint);
             MVM_exception_throw_adhoc(tc, "Error copying one big integer to another: %s", mp_error_to_string(err));
         }
     }
@@ -93,6 +94,7 @@ void MVM_p6bigint_store_as_mp_int(MVMThreadContext *tc, MVMP6bigintBody *body, M
     mp_err err;
     mp_int *i = MVM_malloc(sizeof(mp_int));
     if ((err = mp_init_i64(i, value)) != MP_OKAY) {
+        MVM_free(i);
         MVM_exception_throw_adhoc(tc, "Error creating a big integer from a native integer (%"PRIi64"): %s", value, mp_error_to_string(err));
     }
     body->u.bigint = i;
@@ -129,6 +131,7 @@ static void set_uint(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void
         mp_err err;
         mp_int *i = MVM_malloc(sizeof(mp_int));
         if ((err = mp_init_u64(i, value)) != MP_OKAY) {
+            MVM_free(i);
             MVM_exception_throw_adhoc(tc, "Error creating a big integer from a native integer (%"PRIu64"): %s", value, mp_error_to_string(err));
         }
         body->u.bigint = i;
@@ -207,6 +210,7 @@ static void serialize(MVMThreadContext *tc, MVMSTable *st, void *data, MVMSerial
         }
         buf = (char *)MVM_malloc(len);
         if ((err = mp_to_decimal(i, buf, len)) != MP_OKAY) {
+            MVM_free(buf);
             MVM_exception_throw_adhoc(tc, "Error converting a big integer to a string: %s", mp_error_to_string(err));
         }
 
@@ -239,12 +243,17 @@ static void deserialize(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, vo
         body->u.smallint.value = MVM_serialization_read_int(tc, reader);
     } else {  /* big int */
         mp_err err;
-        char *buf = MVM_string_ascii_encode(tc, MVM_serialization_read_str(tc, reader), NULL, 0);
+        char *buf = MVM_string_ascii_encode_any(tc, MVM_serialization_read_str(tc, reader));
         body->u.bigint = MVM_malloc(sizeof(mp_int));
         if ((err = mp_init(body->u.bigint)) != MP_OKAY) {
+            MVM_free(body->u.bigint);
+            MVM_free(buf);
             MVM_exception_throw_adhoc(tc, "Error initializing a big integer: %s", mp_error_to_string(err));
         }
         if ((err = mp_read_radix(body->u.bigint, buf, 10)) != MP_OKAY) {
+            mp_clear(body->u.bigint);
+            MVM_free(body->u.bigint);
+            MVM_free(buf);
             MVM_exception_throw_adhoc(tc, "Error converting a string to a big integer: %s", mp_error_to_string(err));
         }
         MVM_free(buf);
