@@ -1,13 +1,14 @@
 #include "moar.h"
 
 /* This representation's function pointer table. */
-static const MVMREPROps this_repr;
+static const MVMREPROps P6num_this_repr;
 
 static void mk_storage_spec(MVMThreadContext *tc, MVMuint16 bits, MVMStorageSpec *spec) {
     spec->bits = bits;
     spec->inlineable      = MVM_STORAGE_SPEC_INLINED;
     spec->boxed_primitive = MVM_STORAGE_SPEC_BP_NUM;
     spec->can_box         = MVM_STORAGE_SPEC_CAN_BOX_NUM;
+    spec->is_unsigned     = 0;
     switch (bits) {
         case 64: spec->align = ALIGNOF(MVMnum64); break;
         case 32: spec->align = ALIGNOF(MVMnum32); break;
@@ -18,7 +19,7 @@ static void mk_storage_spec(MVMThreadContext *tc, MVMuint16 bits, MVMStorageSpec
 /* Creates a new type object of this representation, and associates it with
  * the given HOW. */
 static MVMObject * type_object_for(MVMThreadContext *tc, MVMObject *HOW) {
-    MVMSTable *st  = MVM_gc_allocate_stable(tc, &this_repr, HOW);
+    MVMSTable *st  = MVM_gc_allocate_stable(tc, &P6num_this_repr, HOW);
 
     MVMROOT(tc, st, {
         MVMObject *obj = MVM_gc_allocate_type_object(tc, st);
@@ -129,8 +130,11 @@ static void deserialize_repr_data(MVMThreadContext *tc, MVMSTable *st, MVMSerial
     repr_data->bits        = MVM_serialization_read_int(tc, reader);
 
     if (repr_data->bits !=  1 && repr_data->bits !=  2 && repr_data->bits !=  4 && repr_data->bits != 8
-     && repr_data->bits != 16 && repr_data->bits != 32 && repr_data->bits != 64)
-        MVM_exception_throw_adhoc(tc, "MVMP6num: Unsupported int size (%dbit)", repr_data->bits);
+     && repr_data->bits != 16 && repr_data->bits != 32 && repr_data->bits != 64) {
+        MVMint16 bits = repr_data->bits;
+        MVM_free(repr_data);
+        MVM_exception_throw_adhoc(tc, "MVMP6num: Unsupported int size (%dbit)", bits);
+    }
 
     if (repr_data->bits)
         mk_storage_spec(tc, repr_data->bits, &repr_data->storage_spec);
@@ -146,12 +150,64 @@ static void serialize(MVMThreadContext *tc, MVMSTable *st, void *data, MVMSerial
     MVM_serialization_write_num(tc, writer, get_num(tc, st, NULL, data));
 }
 
-/* Initializes the representation. */
-const MVMREPROps * MVMP6num_initialize(MVMThreadContext *tc) {
-    return &this_repr;
+static void spesh(MVMThreadContext *tc, MVMSTable *st, MVMSpeshGraph *g, MVMSpeshBB *bb, MVMSpeshIns *ins) {
+    MVMP6numREPRData *repr_data = (MVMP6numREPRData *)st->REPR_data;
+    switch (ins->info->opcode) {
+        case MVM_OP_box_n: {
+            if (repr_data->bits == 64 && !(st->mode_flags & MVM_FINALIZE_TYPE)) {
+                /* Prepend a fastcreate instruction. */
+                MVMSpeshIns *fastcreate = MVM_spesh_alloc(tc, g, sizeof(MVMSpeshIns));
+                MVMSpeshFacts *tgt_facts = MVM_spesh_get_facts(tc, g, ins->operands[0]);
+                fastcreate->info = MVM_op_get_op(MVM_OP_sp_fastcreate);
+                fastcreate->operands = MVM_spesh_alloc(tc, g, 3 * sizeof(MVMSpeshOperand));
+                fastcreate->operands[0] = ins->operands[0];
+                tgt_facts->writer = fastcreate;
+                fastcreate->operands[1].lit_i16 = st->size;
+                fastcreate->operands[2].lit_i16 = MVM_spesh_add_spesh_slot(tc, g, (MVMCollectable *)st);
+                MVM_spesh_manipulate_insert_ins(tc, bb, ins->prev, fastcreate);
+                tgt_facts->flags |= MVM_SPESH_FACT_KNOWN_TYPE | MVM_SPESH_FACT_CONCRETE;
+                tgt_facts->type = st->WHAT;
+
+                MVM_spesh_graph_add_comment(tc, g, fastcreate, "%s into a %s",
+                        ins->info->name,
+                        MVM_6model_get_stable_debug_name(tc, st));
+
+                /* Change instruction to a bind. */
+                MVM_spesh_usages_delete_by_reg(tc, g, ins->operands[2], ins);
+                ins->info = MVM_op_get_op(MVM_OP_sp_bind_n);
+                ins->operands[2] = ins->operands[1];
+                ins->operands[1].lit_i16 = offsetof(MVMP6num, body.value);
+                MVM_spesh_usages_add_by_reg(tc, g, ins->operands[0], ins);
+            }
+            break;
+        }
+        case MVM_OP_unbox_n:
+        case MVM_OP_decont_n: {
+            if (repr_data->bits == 64 || !repr_data->bits) {
+                /* Lower into a direct memory read. */
+                MVMSpeshOperand *orig_operands = ins->operands;
+
+                MVM_spesh_graph_add_comment(tc, g, ins, "%s from a %s",
+                        ins->info->name,
+                        MVM_6model_get_stable_debug_name(tc, st));
+
+                ins->info = MVM_op_get_op(MVM_OP_sp_get_n);
+                ins->operands = MVM_spesh_alloc(tc, g, 3 * sizeof(MVMSpeshOperand));
+                ins->operands[0] = orig_operands[0];
+                ins->operands[1] = orig_operands[1];
+                ins->operands[2].lit_i16 = offsetof(MVMP6num, body.value);
+            }
+            break;
+        }
+    }
 }
 
-static const MVMREPROps this_repr = {
+/* Initializes the representation. */
+const MVMREPROps * MVMP6num_initialize(MVMThreadContext *tc) {
+    return &P6num_this_repr;
+}
+
+static const MVMREPROps P6num_this_repr = {
     type_object_for,
     MVM_gc_allocate_object,
     NULL, /* initialize */
@@ -184,7 +240,7 @@ static const MVMREPROps this_repr = {
     NULL, /* gc_mark_repr_data */
     gc_free_repr_data,
     compose,
-    NULL, /* spesh */
+    spesh,
     "P6num", /* name */
     MVM_REPR_ID_P6num,
     NULL, /* unmanaged_size */
