@@ -106,7 +106,7 @@ MVMInstance * MVM_vm_create_instance(void) {
     char *spesh_log, *spesh_nodelay, *spesh_disable, *spesh_inline_disable,
          *spesh_osr_disable, *spesh_limit, *spesh_blocking, *spesh_inline_log,
          *spesh_pea_disable;
-    char *jit_expr_disable, *jit_disable, *jit_last_frame, *jit_last_bb;
+    char *jit_expr_enable, *jit_disable, *jit_last_frame, *jit_last_bb;
     char *dynvar_log;
     int init_stat;
 
@@ -155,11 +155,9 @@ MVMInstance * MVM_vm_create_instance(void) {
     init_cond(instance->cond_blocked_can_continue, "GC thread unblock");
 
     /* Safe point free list. */
+    instance->free_at_safepoint = NULL;
     init_mutex(instance->mutex_free_at_safepoint, "safepoint free list");
 
-    /* Create fixed size allocator. */
-    instance->fsa = MVM_fixed_size_create(instance->main_thread);
-
     /* Set up REPR registry mutex. */
     init_mutex(instance->mutex_repr_registry, "REPR registry");
     MVM_index_hash_build(instance->main_thread, &instance->repr_hash, MVM_REPR_CORE_COUNT);
@@ -253,6 +251,8 @@ MVMInstance * MVM_vm_create_instance(void) {
     instance->callsite_interns = MVM_calloc(1, sizeof(MVMCallsiteInterns));
     init_mutex(instance->mutex_callsite_interns, "callsite interns");
 
+    init_mutex(instance->mutex_property_codes_hash_setup, "unicode db lookup hashes");
+
     /* There's some callsites we statically use all over the place. Intern
      * them, so that spesh may end up optimizing more "internal" stuff. */
     MVM_callsite_initialize_common(instance->main_thread);
@@ -312,8 +312,8 @@ MVMInstance * MVM_vm_create_instance(void) {
     if (!jit_disable || !jit_disable[0])
         instance->jit_enabled = 1;
 
-    jit_expr_disable = getenv("MVM_JIT_EXPR_DISABLE");
-    if (!jit_expr_disable || strlen(jit_expr_disable) == 0)
+    jit_expr_enable = getenv("MVM_JIT_EXPR_ENABLE");
+    if (jit_expr_enable && strlen(jit_expr_enable) != 0)
         instance->jit_expr_enabled = 1;
 
 
@@ -656,10 +656,6 @@ void MVM_vm_destroy_instance(MVMInstance *instance) {
     uv_cond_destroy(&instance->cond_blocked_can_continue);
     uv_mutex_destroy(&instance->mutex_gc_orchestrate);
 
-    /* Clean up safepoint free vector. */
-    MVM_VECTOR_DESTROY(instance->free_at_safepoint);
-    uv_mutex_destroy(&instance->mutex_free_at_safepoint);
-
     /* Clean up Hash of HLLConfig. */
     uv_mutex_destroy(&instance->mutex_hllconfigs);
     MVM_fixkey_hash_demolish(instance->main_thread, &instance->compiler_hll_configs);
@@ -678,8 +674,7 @@ void MVM_vm_destroy_instance(MVMInstance *instance) {
     uv_mutex_destroy(&instance->mutex_extop_registry);
     MVM_fixkey_hash_demolish(instance->main_thread, &instance->extop_registry);
 
-    /* Clean up Hash of all known serialization contexts; all SCs list is in
-     * FSA space and so cleaned up with that. */
+    /* Clean up Hash of all known serialization contexts. */
     uv_mutex_destroy(&instance->mutex_sc_registry);
     MVM_str_hash_demolish(instance->main_thread, &instance->sc_weakhash);
 
@@ -702,6 +697,8 @@ void MVM_vm_destroy_instance(MVMInstance *instance) {
     uv_mutex_destroy(&instance->mutex_callsite_interns);
     MVM_callsite_cleanup_interns(instance);
 
+    uv_mutex_destroy(&instance->mutex_property_codes_hash_setup);
+
     /* Clean up syscall registry. */
     MVM_fixkey_hash_demolish(instance->main_thread, &instance->syscalls);
 
@@ -745,13 +742,14 @@ void MVM_vm_destroy_instance(MVMInstance *instance) {
     /* Clean up event loop mutex. */
     uv_mutex_destroy(&instance->mutex_event_loop);
 
+    /* Clean up safepoint free list. */
+    uv_mutex_destroy(&instance->mutex_free_at_safepoint);
+    MVM_alloc_safepoint(instance->main_thread);
+
     /* Destroy main thread contexts and thread list mutex. */
     MVM_tc_destroy(instance->main_thread);
     uv_mutex_destroy(&instance->mutex_threads);
 
-    /* Clean up fixed size allocator */
-    MVM_fixed_size_destroy(instance->fsa);
-
     uv_mutex_destroy(&instance->subscriptions.mutex_event_subscription);
 
     /* Clear up VM instance memory. */
